"""
Description:
"""

# system modules

# django modules
from django.db import models
from django.db.utils import IntegrityError

# model managers
from handyhelpers.managers import HandyHelperModelManager


class HandyHelperBaseModel(models.Model):
    """ abstract model for common fields in models (these fields will appear in all models) """
    objects = HandyHelperModelManager()
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def update(self, **kwargs):
        """ perform an 'update like' operation on a single model instance """
        for i in kwargs:
            setattr(self, i, kwargs[i])
        self.save()

    class Meta:
        abstract = True


class SingletonModel(models.Model):
    """ Singleton model to restrict a database table to one row. """
    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        try:
            self.pk = 1
            super(SingletonModel, self).save(*args, **kwargs)
        except IntegrityError:
            pass

    def delete(self, *args, **kwargs):
        pass

    @classmethod
    def load(cls):
        obj, created = cls.objects.get_or_create(pk=1)
        return obj



from django import template
from django.contrib.auth.models import Group
from django.utils.safestring import mark_safe


register = template.Library()


@register.filter(name='has_group')
def has_group(user, group_name):
    try:
        group = Group.objects.get(name=group_name)
        return group in user.groups.all()
    except:
        return None


@register.filter(name='in')
def inside(value, arg):
    try:
        return value in arg
    except:
        return None


@register.filter
def add(value, arg):
    if arg:
        return value + arg


@register.filter
def subtract(value, arg):
    if arg:
        return value - arg


@register.filter()
def nbsp(value):
    return mark_safe("&nbsp;".join(value.split(' ')))


@register.filter
def intcomma(value):
    if value:
        return "{:,}".format(value)
    else:
        return 0


@register.filter
def usd(value):
    if value:
        return "${0:,.2f}".format(value)
    else:
        return 0


@register.filter
def or_blank(value):
    """ return an empty string if value is None """
    if value:
        return value
    else:
        return ""


@register.filter
def or_zero(value):
    """ return a zero if value is None """
    if value:
        return value
    else:
        return 0


@register.filter(name='percentage')
def percentage(value):
    if not value:
        return ""
    try:
        return "{:4.2f}%".format(float(value)*100)
    except:
        return ""




class FilterByQueryParamsMixin:
    """ Mixin used to evaluate query parameters provided in the URL and update a queryset accordingly. This is typically
    used on list views. Query parameters passed must be valid model fields. Invalid parameters are ignored.

    class parameters:
        request          - request object
        queryset         - django queryset
        page_description - optional parameter used to describe page; typically used as a page subtitle
        distinct         - optional parameter to make queryset include only distinct results

    example usage:
        class HandyHelperGenericBaseListView(FilterByQueryParamsMixin, ListView)
    """

    request = None
    queryset = None
    page_description = None

    def filter_by_query_params(self):
        """
        Description:
            Filters a queryset by query_params in the URL

        Returns:
            filtered queryset
        """
        # pass a description to the view if included as a query parameter
        if not self.page_description and 'page_description' in self.request.GET.dict():
            self.page_description = self.request.GET.dict().get('page_description', None)

        # build a dictionary of valid model fields
        filter_dict = {}
        model = self.queryset.model
        for field, val in self.request.GET.dict().items():
            if field.split("__")[0] not in [i.name for i in model._meta.fields +
                                                            model._meta.many_to_many +
                                                            model._meta.related_objects]:
                continue
            if val is not None:
                if val == 'None':
                    val = None
                filter_dict[field] = val
        if 'distinct' in self.request.GET.dict():
            return self.queryset.filter(**filter_dict).distinct()
        return self.queryset.filter(**filter_dict)



"""
DESCRIPTION:
    class definitions for common model managers
"""

# system modules
from random import randint

# django models
from django.db import models


class HandyHelperModelManager(models.Manager):
    def get_object_or_none(self, **kwargs):
        """ return object if available; return None if not available """
        try:
            return super(HandyHelperModelManager, self).get(**kwargs)
        except models.ObjectDoesNotExist:
            return None

    def filter_or_none(self, **kwargs):
        """ return object if available; return None if not available """
        try:
            return super(HandyHelperModelManager, self).filter(**kwargs)
        except models.ObjectDoesNotExist:
            return None

    def get_latest_or_none(self, *args, **kwargs):
        """ return latest object id available; None if not available """
        try:
            return super(HandyHelperModelManager, self).latest(*args, **kwargs)
        except models.ObjectDoesNotExist:
            return None

    def get_fields(self, exclude_list=('OneToOneField')):
        """ return a list of fields in the model """
        if not issubclass(self.model, models.Model):
            return []
        return [i for i in self.model._meta.fields if type(i).__name__ not in exclude_list]

    def get_field_names(self, exclude_list=('OneToOneField')):
        """ return a list of field names in the model """
        if not issubclass(self.model, models.Model):
            return []
        return [i.name for i in self.model._meta.fields if type(i).__name__ not in exclude_list]

    def get_properties(self):
        """ return a list of model property names """
        if not issubclass(self.model, models.Model):
            return []
        properties_list = []
        for name in dir(self.model):
            if isinstance(getattr(self.model, name), property):
                properties_list.append(name)
        return properties_list

    def get_fields_and_properties(self):
        """ return a list of fields and properties in a model """
        return self.get_fields() + self.get_properties()

    def get_foreign_keys(self):
        """ return a list of foreignKeys in the model """
        if not issubclass(self.model, models.Model):
            return []
        return [i for i in self.model._meta.fields if i.get_internal_type() == "ForeignKey"]

    def get_foreign_key_names(self):
        """ return a list of names of foreignKeyfields """
        return [i.name for i in self.get_foreign_keys()]

    def get_random_row(self, **kwargs):
        """ return a single, random entry from a queryset """
        queryset = self.filter(**kwargs)
        if queryset:
            random_index = randint(0, queryset.count() - 1)
            return queryset[random_index]
        return None


class ParentModelMixin(object):
    """ methods for parent models """

    def get_child_list(self):
        """ return a list of child objects """
        return [f.get_accessor_name() for f in self._meta.get_fields()
                if getattr(f, 'field_name', None) and f.one_to_one]

    def get_child(self):
        """ return the child inherited from this parent object instance """
        for i in self.get_child_list():
            child = getattr(self, i, None)
            if child:
                return child
        return None

    def get_grandchild(self):
        """ return the grandchild inherited from this parent object instance """
        child = self.get_child()
        if child:
            return child.get_child()
        return None



"""
Description:
    Collection of helpers to assist with queryset operations

"""

# import system modules
import datetime
import collections

# import Django modules
from django.db.models import Count
from django.utils import timezone


def count_by_hour(queryset, field_name):
    """
    sort queryset results by hour of day

    Args:
        queryset: django queryset
        field_name: field to group data by (string)

    Returns:
        list of grouped values: count of entries per hour, starting with current hour, descending chronologically
    """
    data = queryset.extra({'hour': "hour({})".format(field_name)}).values('hour').annotate(Count('id'))
    new_dict = collections.OrderedDict()
    for i in data:
        values = i.values()
        new_dict[values[0]] = values[1]

    return_list = []
    now = timezone.now()
    for i in [(now - datetime.timedelta(hours=i)).hour for i in range(0, 24)]:
        if i not in new_dict.keys():
            return_list.append(0)
        else:
            return_list.append(new_dict[i])
    return return_list


def count_by_week(queryset, field_name):
    """
    Description:
        sort queryset results by week of year

    Args:
        queryset: django queryset
        field_name: field to group data by (string)

    Returns:
        list of grouped values
        list of grouped values: count of entries per week, starting with current week, descending chronologically
    """
    data = queryset.extra({'week': "week({})".format(field_name)}).values('week').annotate(Count('id'))
    new_dict = collections.OrderedDict()
    for i in data:
        new_dict[i['week']] = i['id__count']

    return_list = []
    now = timezone.now()
    for i in [(now - datetime.timedelta(weeks=i)).isocalendar()[1] for i in range(0, 52)]:
        if i not in new_dict.keys():
            return_list.append(0)
        else:
            return_list.append(new_dict[i])
    return return_list


def count_by_month(queryset, field_name):
    """
    Description:
        sort queryset results by month of year

    Args:
        queryset: django queryset
        field_name: field to group data by (string)

    Returns:
        list of grouped values: count of entries per month, starting with current month, descending chronologically
    """
    data = queryset.extra({'month': "month({})".format(field_name)}).values('month').annotate(Count('id'))
    new_dict = collections.OrderedDict()
    for i in data:
        new_dict[i['month']] = i['id__count']

    return_list = []
    now = timezone.now()
    for i in [(now - datetime.timedelta(i * 365 / 12)).month for i in range(0, 12)]:
        if i not in new_dict.keys():
            return_list.append(0)
        else:
            return_list.append(new_dict[i])
    return return_list



"""
This file contains custom filter lookups as per: https://docs.djangoproject.com/en/2.2/howto/custom-lookups/

https://docs.djangoproject.com/en/2.2/ref/models/lookups/
https://docs.djangoproject.com/en/2.2/ref/models/querysets/#field-lookups
"""

from django.db.models import Lookup


class NotEqual(Lookup):
    lookup_name = 'ne'

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return '%s <> %s' % (lhs, rhs), params


class Exclude(Lookup):
    lookup_name = 'excludes'

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return 'NOT (%s LIKE %s)' % (lhs, rhs), params



"""
This file contains view that perform an action and are redirected to another URL. These view do not render a template.
"""

from django.shortcuts import redirect
from django.views.generic import (View)


class FilterListView(View):
    """ apply filters, as provided via queryparameters, to a list view that uses the FilterByQueryParamsMixin """
    def post(self, request, *args, **kwargs):
        """ process POST request """
        redirect_url = self.request.META.get('HTTP_REFERER')
        form_parameters = self.request.POST.dict()

        # remove csrf token from POST parameters
        if form_parameters.get('csrfmiddlewaretoken'):
            form_parameters.pop('csrfmiddlewaretoken')

        # build filtered URL
        filter_url = f'{redirect_url.split("/?")[0]}?'
        for key, value in form_parameters.items():
            if value:
                filter_url += f'{key}={value}&'

        return redirect(filter_url)


class ShowAllListView(View):
    """ show all records (undo a FilterListView) """
    def get(self, request, *args, **kwargs):
        """ process GET request """
        redirect_url = self.request.META.get('HTTP_REFERER')
        return redirect(redirect_url.split("/?")[0])



from django.urls import path
from handyhelpers.views import action

app_name = 'handyhelpers'

urlpatterns = [
    # action views
    path('filter_list_view', action.FilterListView.as_view(), name='filter_list_view'),
    path('show_all_list_view', action.ShowAllListView.as_view(), name='show_all_list_view'),
]



"""
django-handy-helpers library
----------------------------
A collection of handy utilities to support django operations.
:copyright: (c) 2018 by David Slusser
:license: GPL-3.0. See LICENSE for more details
"""

__title__ = 'django-handyhelpers'
__version__ = '0.0.28'
__author__ = 'David Slusser'
__email__ = 'dbslusser@gmail.com'
__license__ = 'GPL-3.0'
__copyright__ = 'Copyright 2020 David Slusser'

default_app_config = 'handyhelpers.apps.HandyHelpersConfig'



from django.conf import settings
from django.shortcuts import render
from django.views.generic import ListView, View

from handyhelpers.mixins.view_mixins import FilterByQueryParamsMixin


class HandyHelperGenericBaseView(View):
    """ Generic view used to set the base_template variable. This extends django.view.generic.View and will use the base
     template defined in the BASE_TEMPLATE settings variable or use handyhelpers/handyhelpers_base.htm if not provided.

    class parameters:
        base_template - base template used for rendering page; defaults to: handyhelpers_base.htm
        args          - additional args to pass into the template
        kwargs        - additional kwargs to pass into the template
    """
    base_template = getattr(settings, 'BASE_TEMPLATE', 'handyhelpers/handyhelpers_base.htm')
    args = None
    kwargs = None


class HandyHelperGenericBaseItemizedView(HandyHelperGenericBaseView):
    """ Generic view to include shared variables used in views itemized views, such as index and action views.

    class parameters:
        base_template        - base template used for rendering page; defaults to: handyhelpers_base.htm
        args                 - additional args to pass into the template
        kwargs               - additional kwargs to pass into the template
        title                - page title to use; will display in a H1 element
        subtitle             - subtitle to use; will display in a H5 element (None if not provided)
        item_list            - list of items/actions to display (one per card)
        protected_item_list  - list of items/action only visible to members of protected_group_name
        protected_group_name - name of group to show protected_item_list items to
    """
    title = None
    subtitle = None
    item_list = None
    protected_item_list = None
    protected_group_name = None


class HandyHelperGenericBaseListView(FilterByQueryParamsMixin, ListView):
    """ Generic list view used to set the base_template, template, title, table, and modal variables. This extends
    django.view.generic.ListView and will use the base template defined in the BASE_TEMPLATE settings variable or use
    handyhelpers/handyhelpers_base.htm if not provided. This view also includes the
    handyhelpers.mixins.view_mixins.FilterByQueryParamsMixin mixin to allow filtering by query parameters.

    class parameters:
        base_template - base template used for rendering page; defaults to: handyhelpers_base.htm
        template_name - template used when rendering page; defaults to: handyhelpers/generic/generic_list.html
        title         - title to use in template
        table         - htm file rendering the queryset to be included in the generic_list template
        modals        - htm file rendering additional modals to be included in the generic_list template
    """
    base_template = getattr(settings, 'BASE_TEMPLATE', 'handyhelpers/handyhelpers_base.htm')
    template_name = 'handyhelpers/generic/generic_list.html'
    title = None
    table = None
    modals = None
    args = None
    kwargs = None


class HandyHelperIndexView(HandyHelperGenericBaseItemizedView):
    """
    View to render an index page with actions/options listed as 'zoom-cards' in flex-box container(s)

    class parameters:
        base_template        - base template used for rendering page; defaults to: handyhelpers_base.htm
        template_name        - template used when rendering page; defaults to: handyhelpers/generic/generic_index.html
        title                - page title to use; will display in a H1 element
        subtitle             - subtitle to use; will display in a H5 element (None if not provided)
        item_list            - list of items/actions to display (one per card)
        protected_item_list  - list of items/action only visible to members of protected_group_name
        protected_group_name - name of group to show protected_item_list items to
    """
    template_name = 'handyhelpers/generic/generic_index.html'

    def get(self, request):
        context = dict(base_template=self.base_template, title=self.title, subtitle=self.subtitle,
                       item_list=self.item_list, protected_item_list=self.protected_item_list,
                       protected_group_name=self.protected_group_name, args=self.args, kwargs=self.kwargs)
        return render(request, self.template_name, context)


class HandyHelperActionView(HandyHelperGenericBaseItemizedView):
    """
    Action page with actions/options listed as 'zoom-cards' in flex-box container(s)

    class parameters:
        base_template        - base template used for rendering page; defaults to: handyhelpers_base.htm
        args                 - additional args to pass into the template
        kwargs               - additional kwargs to pass into the template
        template_name        - template used when rendering page; defaults to: handyhelpers/generic/generic_action.html
        title                - page title to use; will display in a H1 element
        subtitle             - subtitle to use; will display in a H5 element (None if not provided)
        item_list            - list of items/actions to display (one per card)
        protected_item_list  - list of items/action only visible to members of protected_group_name
        protected_group_name - name of group to show protected_item_list items to
        form_list            - list of django forms used in page
    """
    template_name = 'handyhelpers/generic/generic_action.html'
    form_list = None

    def get(self, request):
        context = dict(base_template=self.base_template, title=self.title, subtitle=self.subtitle,
                       item_list=self.item_list, form_list=self.form_list, protected_item_list=self.protected_item_list,
                       protected_group_name=self.protected_group_name, args=self.args, kwargs=self.kwargs)
        return render(request, self.template_name, context)


class HandyHelperAboutView(HandyHelperGenericBaseView):
    """
    Generic view to render an 'about' page.

    class parameters:
        base_template - base template used for rendering page; defaults to: handyhelpers_base.htm
        args          - additional args to pass into the template
        kwargs        - additional kwargs to pass into the template
        template_name - template used when rendering page; defaults to: handyhelpers/generic/generic_about.html
        title         - page title to use ('About' if not provided)
        subtitle      - subtitle to use (None if not provided)
        version       - version to display (None if not provided)
        details       - any specific details to display (None if not provided)
        source        - link to source code repository (None if not provided)
        contact       - any desired contact information (None if not provided)
        links         - list of dictionary containing links; ex. [{'some reference': 'www.somewebsite.com}, ...]
    """
    template_name = 'handyhelpers/generic/generic_about.html'
    title = 'About'
    subtitle = None
    version = None
    details = None
    source = None
    contact = None
    links = []

    def get(self, request):
        context = dict(base_template=self.base_template, title=self.title, subtitle=self.subtitle, version=self.version,
                       details=self.details, source=self.source, contact=self.contact, links=self.links,
                       args=self.args, kwargs=self.kwargs)
        return render(request, self.template_name, context)


class HandyHelperSingletonView(HandyHelperGenericBaseView):
    """
    View for rendering a page based on a SingletonModel

    class parameters:
        base_template - base template used for rendering page; defaults to: handyhelpers_base.htm
        args          - additional args to pass into the template
        kwargs        - additional kwargs to pass into the template
        template_name - template used when rendering page
        model         - singleton model
    """
    template_name = None
    model = None

    def get(self, request):
        context = dict(object=self.model.objects.get(), args=self.args, kwargs=self.kwargs)
        return render(request, self.template_name, context)


class HandyHelperListView(HandyHelperGenericBaseListView):
    """
    A reusable generic base view to render a ListView where the child view will provide a html table.

    class parameters:
        base_template    - base template used for rendering page; defaults to: handyhelpers_base.htm
        template_name    - template used when rendering page; defaults to: handyhelpers/generic/generic_list.html
        args             - additional args to pass into the template
        kwargs           - additional kwargs to pass into the template
        queryset         - queryset to be rendered on the page
        title            - title to use in template
        page_description - subtitle to use in template
        table            - htm file rendering the queryset to be included in the generic_list template
        modals           - htm file rendering additional modals to be included in the generic_list template

    example:
        class ListProjects(HandyHelperBaseListView):
            queryset = Project.objects.all()
            title = 'Projects'
            page_description = 'my cool projects'
            table = 'table/table_projects.htm'
            modals = 'project_modals.htm'
    """
    def get(self, request, *args, **kwargs):
        context = dict(base_template=self.base_template, queryset=self.filter_by_query_params(), title=self.title,
                       sub_title=self.page_description, table=self.table, modals=self.modals,
                       args=self.args, kwargs=self.kwargs)
        return render(request, self.template_name, context)


class HandyHelperBaseListView(HandyHelperListView):
    """ maintaining the HandyHelperBaseListView for legacy compatibility """
    pass


class HandyHelperListPlusCreateView(HandyHelperGenericBaseListView):
    """
    A reusable generic base view to render a ListView where the child view will provide a html table and create form.

    class parameters:
        base_template          - base template used for rendering page; defaults to: handyhelpers_base.htm
        template_name          - template used when rendering page; defaults to: handyhelpers/generic/generic_list.html
        args                   - additional args to pass into the template
        kwargs                 - additional kwargs to pass into the template
        queryset               - queryset to be rendered on the page
        title                  - title to use in template
        page_description       - subtitle to use in template
        table                  - htm file rendering the queryset to be included in the generic_list template
        modals                 - htm file rendering additional modals to be included in the generic_list template

        create_form_obj        - form object
        create_form_url        - url the form (action) should post to
        create_form_title      - title to use on the create form modal (can be html)
        create_form_modal      - name of modal for the create form
        create_form_link_title - text used for the link opening the create form
        create_form_tool_tip   - text to use for the tooltip

    example:
        class ListProjects(HandyHelperBaseCreateListView):
            queryset = Project.objects.all()
            title = 'Projects'
            page_description = 'my cool projects'
            table = 'table/table_projects.htm'

            create_form_obj = ProjectForm
            create_form_url = '/hostmgr/create_project/'
            create_form_title = '<b>Add Project: </b><small> </small>'
            create_form_modal = 'add_project'
            create_form_link_title = 'add project'
            create_form_tool_tip = 'add project'
    """
    create_form = dict()
    create_form_obj = None
    create_form_id = None
    create_form_url = None
    create_form_title = None
    create_form_modal = None
    create_form_link_title = None
    create_form_tool_tip = None

    def get(self, request, *args, **kwargs):
        context = dict(base_template=self.base_template, queryset=self.filter_by_query_params(), title=self.title,
                       sub_title=self.page_description, table=self.table, modals=self.modals,
                       args=self.args, kwargs=self.kwargs)
        if self.create_form_obj:
            self.create_form['form'] = self.create_form_obj(request.POST or None)
            self.create_form['form_id'] = self.create_form_id
            self.create_form['action'] = 'Add'
            self.create_form['action_url'] = self.create_form_url
            self.create_form['title'] = self.create_form_title
            self.create_form['modal_name'] = self.create_form_modal
            self.create_form['link_title'] = self.create_form_link_title
            self.create_form['tool_tip'] = self.create_form_tool_tip
            context['create_form'] = self.create_form
        return render(request, self.template_name, context)


class HandyHelperBaseListPlusCreateView(HandyHelperListView):
    """ maintaining the HandyHelperBaseListPlusCreateView for legacy compatibility """
    pass


class HandyHelperListPlusFilterView(HandyHelperGenericBaseListView):
    """
    A reusable generic base view to render a ListView where the child view will provide a html table and filter form.

    class parameters:
        base_template          - base template used for rendering page; defaults to: handyhelpers_base.htm
        template_name          - template used when rendering page; defaults to: handyhelpers/generic/generic_list.html
        args                   - additional args to pass into the template
        kwargs                 - additional kwargs to pass into the template
        queryset               - queryset to be rendered on the page
        title                  - title to use in template
        page_description       - subtitle to use in template
        table                  - htm file rendering the queryset to be included in the generic_list template
        modals                 - htm file rendering additional modals to be included in the generic_list template

        filter_form_obj        - form object
        filter_form_url        - url the form (action) should post to
        filter_form_title      - title to use on the filter form modal (can be html)
        filter_form_modal      - name of modal for the filter form
        filter_form_link_title - text used for the link opening the filter form
        filter_form_tool_tip   - text to use for the tooltip
        filter_form_undo = True

    example:
        class ListProjects(HandyHelperListPlusFilterView):
            queryset = Project.objects.all()
            title = 'Projects'
            page_description = 'my cool projects'
            table = 'table/table_projects.htm'

            filter_form_obj = ProjectForm
            filter_form_url = '/hostmgr/filter_project/'
            filter_form_title = '<b>Filter Projects: </b><small> </small>'
            filter_form_modal = 'filter_project'
            filter_form_link_title = 'filter project'
            filter_form_tool_tip = 'filter project'
    """
    filter_form = dict()
    filter_form_obj = None
    filter_form_id = None
    filter_form_url = '/handyhelpers/filter_list_view'
    filter_form_title = None
    filter_form_modal = None
    filter_form_link_title = None
    filter_form_tool_tip = None
    filter_form_undo = True

    def get(self, request, *args, **kwargs):
        context = dict(base_template=self.base_template, queryset=self.filter_by_query_params(), title=self.title,
                       sub_title=self.page_description, table=self.table, modals=self.modals,
                       args=self.args, kwargs=self.kwargs)
        if self.filter_form_obj:
            self.filter_form['form'] = self.filter_form_obj(request.POST or None)
            self.filter_form['form_id'] = self.filter_form_id
            self.filter_form['action'] = 'Filter'
            self.filter_form['action_url'] = self.filter_form_url
            self.filter_form['title'] = self.filter_form_title
            self.filter_form['modal_name'] = self.filter_form_modal
            self.filter_form['link_title'] = self.filter_form_link_title
            self.filter_form['tool_tip'] = self.filter_form_tool_tip
            self.filter_form['undo'] = self.filter_form_undo
            context['filter_form'] = self.filter_form
        return render(request, self.template_name, context)


class HandyHelperListPlusCreateAndFilterView(HandyHelperGenericBaseListView):
    """
    A reusable generic base view to render a ListView where the child view will provide a html table and
    and a create form and a filter form.

    class parameters:
        base_template          - base template used for rendering page; defaults to: handyhelpers_base.htm
        template_name          - template used when rendering page; defaults to: handyhelpers/generic/generic_list.html
        args                   - additional args to pass into the template
        kwargs                 - additional kwargs to pass into the template
        queryset               - queryset to be rendered on the page
        title                  - title to use in template
        page_description       - subtitle to use in template
        table                  - htm file rendering the queryset to be included in the generic_list template
        modals                 - htm file rendering additional modals to be included in the generic_list template

        create_form_obj        - create form object
        create_form_url        - url the create form (action) should post to
        create_form_title      - title to use on the create form modal (can be html)
        create_form_modal      - name of modal for the create form
        create_form_link_title - text used for the link opening the create form
        create_form_tool_tip   - text to use for the create form link tooltip

        filter_form_obj        - filter form object
        filter_form_url        - url the filter form (action) should post to
        filter_form_title      - title to use on the filter form modal (can be html)
        filter_form_modal      - name of modal for the filter form
        filter_form_link_title - text used for the link opening the filter form
        filter_form_tool_tip   - text to use for the filter form link tooltip
        filter_form_undo       - include an undo icon to clear applied filters

    example:
        class ListProjects(HandyHelperBaseCreateListView):
            queryset = Project.objects.all()
            title = 'Projects'
            page_description = 'my cool projects'
            table = 'table/table_projects.htm'

            create_form_obj = ProjectForm
            create_form_url = '/hostmgr/create_project/'
            create_form_title = '<b>Add Project: </b><small> </small>'
            create_form_modal = 'add_project'
            create_form_link_title = 'add project'
            create_form_tool_tip = 'add project'

            filter_form_obj = ProjectForm
            filter_form_url = '/hostmgr/filter_project/'
            filter_form_title = '<b>Filter Projects: </b><small> </small>'
            filter_form_modal = 'filter_projects'
            filter_form_link_title = 'filter projects'
            filter_form_tool_tip = 'filter projects'
            filter_form_undo = True
    """
    create_form = dict()
    create_form_obj = None
    create_form_url = None
    create_form_title = None
    create_form_modal = None
    create_form_link_title = None
    create_form_tool_tip = None

    filter_form = dict()
    filter_form_obj = None
    filter_form_url = '/handyhelpers/filter_list_view'
    filter_form_title = None
    filter_form_modal = None
    filter_form_link_title = None
    filter_form_tool_tip = None
    filter_form_undo = True

    def get(self, request, *args, **kwargs):
        context = dict(base_template=self.base_template, queryset=self.filter_by_query_params(), title=self.title,
                       sub_title=self.page_description, table=self.table, modals=self.modals,
                       args=self.args, kwargs=self.kwargs)
        if self.create_form_obj:
            self.create_form['form'] = self.create_form_obj(request.POST or None)
            self.create_form['action'] = 'Add'
            self.create_form['action_url'] = self.create_form_url
            self.create_form['title'] = self.create_form_title
            self.create_form['modal_name'] = self.create_form_modal
            self.create_form['link_title'] = self.create_form_link_title
            self.create_form['tool_tip'] = self.create_form_tool_tip
            context['create_form'] = self.create_form

        if self.filter_form_obj:
            self.filter_form['form'] = self.filter_form_obj(request.POST or None)
            self.filter_form['action'] = 'Filter'
            self.filter_form['action_url'] = self.filter_form_url
            self.filter_form['title'] = self.filter_form_title
            self.filter_form['modal_name'] = self.filter_form_modal
            self.filter_form['link_title'] = self.filter_form_link_title
            self.filter_form['tool_tip'] = self.filter_form_tool_tip
            self.filter_form['undo'] = self.filter_form_undo
            context['filter_form'] = self.filter_form

        return render(request, self.template_name, context)
